local version = "PolyPrinter post-processing v009"
print( version )
local x = os.clock()

-- TODO: use cooling fan setting as the minimum setting, go to 100% always at the real min time
-- add the min time fraction as a commentline with name = value pair parsing into a table automatically for all 
--   (stop loooking after the initial section)
-- stop looking for some of the patterns once we're past where they would be found
-- 005 don't eat Z, E jerk settings
-- 006 clean up acceleration handling
-- 007, 008 better jerk and accel handling, defaulting



local replaceOriginal = true 

local bDoJerkModification = true

BEGIN_LAYER_PATTERN  = "^G92 E0";
PERIMETER_PATTERN  = "^; \'Perimeter\'";
LOOP_PATTERN       = "^; \'Loop\'";
SPARSEFILL_PATTERN = "^; \'Sparse Infill\'";  -- hopefully the first layer above stacked sparse?

MOVE4_PATTERN	   		= "^G1 X.+Y.+Z.+E"  -- fast match of basic possibility of a 4-axis move pattern
MOVE4_PARSE_PATTERN	    = "^G1 X([-%d%.]+)%s+Y([-%d%.]+)%s+Z([-%d%.]+)%s+E([-%d%.]+)" -- suitable for a full parse

MOVEXYEF_PATTERN	   		= "^G1 X.+Y.+E.+F"  -- fast match of basic possibility
MOVEXYEF_PARSE_PATTERN	    = "^G1 X([-%d%.]+)%s+Y([-%d%.]+)%s+E([-%d%.]+)%s+F([-%d%.]+)" -- suitable for a full parse

MOVE3_PATTERN	   		= "^G1 X.+Y.+E"  -- fast match of basic possibility of a move pattern
MOVE3_PARSE_PATTERN	    = "^G1 X([-%d%.]+)%s+Y([-%d%.]+)%s+E([-%d%.]+)" -- suitable for a full parse

MOVE_E_PATTERN			= "^G1 E"  -- fast match of basic possibility of an extruder move pattern
MOVE_E_PARSE_PATTERN	= "^G1 E([-%d%.]+)" -- suitable for a full parse

-- ; *** Slowing to match Min Layer Time (speed multiplier is 0.079412) ***
SLOWING_DOWN_PATTERN = "^; *** Slowing"  -- fast match of basic possibility 
SLOWING_DOWN_PARSE_PATTERN	= "speed multiplier is ([-%d%.]+)" -- suitable for a full parse

-- last known axis positions
local prevX = 0
local prevY = 0
local prevZ = 0
local prevE = 0

function clamp( num, lowLimit, hiLimit )
	num = math.min( num, hiLimit )
	return math.max( num, lowLimit )
end

function linearInterpolatePct( fpct, minVal, maxVal )
	if ( minVal == nil ) then return maxVal end
	if ( maxVal == nil ) then return minVal end
	return minVal + fpct * ( maxVal - minVal );
end

function linearInterpolatePct_Clamp( fpct, minVal, maxVal )
	fpct = math.min( fpct, 1.0 )
	fpct = math.max( fpct, 0 )
	return minVal + fpct * ( maxVal - minVal );
end


function getNumericSettingValue( assignmentLine )
	local dummy;
	local pattern = "([%w ]+)=%s*([%d%.]+)"
	local key, value
	dummy, dummy, key, value = assignmentLine:find( pattern )
	--print( "key " .. key .. " value " .. value )
	return tonumber( value );
end

-- parse a GCODE line of the form: M205 X05
-- return M, 205, X, 205 all separated out
function parseGCodeSetting1Value( gcodeLine )
	local dummy;
	local pattern = "(%a)(%d+)%s*(%a)([-%d%.]+)"
	local gtype, codeNum, axis1, axis1val
	dummy, dummy, gtype, codeNum, axis1, axis1val = gcodeLine:find( pattern )
	--print( "key " .. gtype .. " value " .. codeNum )
	return gtype, tonumber( codeNum ), axis1, tonumber( axis1val )
end


-- parse a GCODE line of the form: M201 X3000 Y3000
-- return M, 201, X, 3000, Y, 3000 all separated out
function parseGCodeSetting2Value( gcodeLine )
	--print( "parseGCodeSetting2Value" .. gcodeLine )

	local dummy;
	local pattern = "(%a)(%d+)%s*(%a)([-%d%.]+)%s*(%a)([-%d%.]+)"
	local gtype, codeNum, axis1, axis1val, axis2, axis2val
	dummy, dummy, gtype, codeNum, axis1, axis1val, axis2, axis2val = gcodeLine:find( pattern )
	--print( "key " .. gtype .. " value " .. codeNum )
	return gtype, tonumber( codeNum ), axis1, tonumber( axis1val ), axis2, tonumber( axis2val )
end

-- should only be called when line is known to use G01 X... Y... Z... E...
function parseGCodeMove4( gcodeLine )
	--print( "parseGCodeMove4" .. gcodeLine )

	local dummy;
	local pattern = MOVE4_PARSE_PATTERN
	local axisX, axisY, axisZ, axisE
	dummy, dummy, axisX, axisY, axisZ, axisE = gcodeLine:find( pattern )
	--print( "X " .. axisX .. " Y " .. axisY .. " Z " .. axisZ .. " E " .. axisE )
	return tonumber( axisX ), tonumber( axisY ), tonumber( axisZ ), tonumber( axisE )
end


-- should only be called when line is known to use G01 X... Y... E...
function parseGCodeMove3( gcodeLine )
	--print( "parseGCodeMove3" .. gcodeLine )

	local dummy;
	local pattern = MOVE3_PARSE_PATTERN
	local axisX, axisY, axisE
	dummy, dummy, axisX, axisY, axisE = gcodeLine:find( pattern )
	--print( "X " .. axisX .. " Y " .. axisY .. " E " .. axisE )
	return tonumber( axisX ), tonumber( axisY ), tonumber( axisE )
end

-- should only be called when line is known to use G01 X... Y... E... F...
function parseGCodeMoveXYEF( gcodeLine )
	--print( "parseGCodeMoveXYEF" .. gcodeLine )
	
	local pattern = MOVEXYEF_PARSE_PATTERN
	local dummy, dummy2, axisX, axisY, axisE, feedrate = gcodeLine:find( MOVEXYEF_PARSE_PATTERN )
	--print( "X " .. axisX .. " Y " .. axisY .. " E " .. axisE .. " F " .. feedrate )
	return tonumber( axisX ), tonumber( axisY ), tonumber( axisE ), tonumber( feedrate )
end

-- should only be called when line is known to use G01 E...
function parseGCodeMoveE( gcodeLine )
	--print( "parseGCodeMoveE" .. gcodeLine )

	local dummy;
	local pattern = MOVE_E_PARSE_PATTERN
	local axisE
	dummy, dummy, axisE = gcodeLine:find( pattern )
	--print( " E " .. axisE )
	return tonumber( axisE )
end

function parseSlowdownFactor( gcodeLine )
	local dummy;
	local factor
	dummy, dummy, factor = gcodeLine:find( SLOWING_DOWN_PARSE_PATTERN )
	--print( " slowdown ", factor, " Z ", prevZ )
	return tonumber( factor )
end

-- parse a GCODE line of the form: M201 X3000 Y3000 Znnn where any of them may be missing or re-ordered
-- returns the words that make up the line
function parseGCodeWords( gcodeLine )

	local words = {}
	-- TODO: include all possible code letters
	--            only read up to comment indicator
	for word in line:gmatch( "[A-Z][-%d.]+" ) do table.insert( words, word ) end

	return words;
end

function makeGCode4Value( type, codeNum, axis1, axis1val, axis2, axis2val, axis3, axis3val, axis4, axis4val )
	return table.concat( { type, codeNum, " " , axis1, axis1val, " ", axis2, axis2val, " ", axis3, axis3val, " ", axis4, axis4val } ) 
end

function makeGCode2Value( type, codeNum, axis1, axis1val, axis2, axis2val )
	--return type .. codeNum .. " " .. axis1 .. axis1val .. " " .. axis2 .. axis2val
	return table.concat( { type, codeNum, " " , axis1, axis1val, " ", axis2, axis2val } ) 
end

function makeGCode1Value( type, codeNum, axis1, axis1val )
	return type .. codeNum .. " " .. axis1 .. axis1val
end

print( arg[1] )
print( inFileName )
local inFileName  = arg[1]
local outFileName = ( inFileName .. "post.gcode" )

local fin  = assert( io.open( inFileName ) )

--local fout = assert( io.open( outFileName, "wb" ) )  -- binary mode saves a few characters - but it's not easily editable by Notepad then
local fout = assert( io.open( outFileName, "w" ) )

fout:setvbuf("full", 2^24)
fout:write( "; ", version , '\n' )

-- Vary the acceleration and jerk settings with quality setting
-- these can probably be extracted from the Prefix Gcode. It could have two
-- sets of settings, understood to encompass the range of quality settings

local DEFAULT_MAX_ACCEL = 3000
local DEFAULT_MIN_ACCEL = 500
local DEFAULT_MAX_JERK = 25
local DEFAULT_MIN_JERK =  5

-- default these to nil for easy managmentof the different possibilities
local maxAccelX = nil
local maxAccelY = nil

local minAccelX = nil
local minAccelY = nil

local maxJerkXY = nil
local minJerkXY = nil

local quality = 1
local foundAccelSpec = false
local foundJerkSpec = false
local inMainGCode = false
local didModifyAccel = false			-- will be true if the last settings were not the default


-- strings are unique in Lua, so they serve as unique, fast identifiers
-- states:
local stateNull = nil
local stateEmitAccelOnExtrudeMove = "stateEmitAccelOnExtrudeMove"
local state = stateNull

local slowdownstate = nil
local slowdownState_FixUpFan = "slowdownState_FixUpFan"


local currentAccelX = maxAccelX
local currentAccelY = maxAccelY
local currentJerkXY = maxJerkXY

function writeCurrentAccelerationLine( ofile )
	ofile:write( makeGCode2Value( "M", 201, "X", currentAccelX, "Y", currentAccelY ) , "\n" )
end

function writeCurrentJerkLine( ofile )
	if ( bDoJerkModification ) then
		ofile:write( makeGCode1Value( "M", 205, "X", currentJerkXY ), "\n" )
	end
end

function writeGivenXYEFLine( ofile, x, y, e, f )
	ofile:write( makeGCode4Value( "G", 1, "X", x, "Y", y, "E", e, "F", f ) , "\n" )
end

function writeFanControlLine( ofile, fan )
	ofile:write( makeGCode1Value( "M", 106, "S", fan ), "\n" )
end


--[[
There are lots of different sections in the Kss slicer output:
; 'Skirt', 4.4 [feed mm/s], 60.0 [head mm/s]
; 'Wipe (and De-string)', 0.0 [feed mm/s], 60.0 [head mm/s]
; 'Solid', 2.2 [feed mm/s], 60.0 [head mm/s]
; 'Stacked Sparse Infill', 6.3 [feed mm/s], 151.7 [head mm/s]

The most important one for us to modify is:
; 'Perimeter', 2.2 [feed mm/s], 60.0 [head mm/s]

Fairly important:
; 'Loop', 2.2 [feed mm/s], 60.0 [head mm/s]
; 'Crown', 2.6 [feed mm/s], 91.2 [head mm/s]


Sections appear to end with a line with a single ';' so we can probably use that
to restore the default settings at the end of the sections we modify.
Alternatively, we set the accelerations every time any section begins.
]]--
local lastPctFast = -1

function writeAccelSettings( pctFast )
	pctFast = clamp( pctFast, 0, 1.0 )
	if ( pctFast ~= lastPctFast ) then
		-- write out a line that sets it, based on the quality setting
		currentAccelX = math.ceil( linearInterpolatePct( pctFast, minAccelX, maxAccelX ) )
		currentAccelY = math.ceil( linearInterpolatePct( pctFast, minAccelY, maxAccelY ) )
		currentJerkXY = math.ceil( linearInterpolatePct( pctFast, minJerkXY, maxJerkXY ) )
		
		-- acceleration and jerk sttings take place immediately, no buffered like Gcode.
		-- simplest is to drain the buffer first.
		-- NOTE: this causes a stall doing the last action because the planner always executes it at slow speed, hoping for more
		--             - set a flag, or drain to 1 left, or handle accel changes in synch (pick up when executed)
		-- not good enough by itself fout:write( "M400\n" )
		
		writeCurrentAccelerationLine( fout )
		writeCurrentJerkLine( fout )
		didModifyAccel = ( pctFast < 1.0 )
		lastPctFast = pctFast
	end
	--if ( pctFast >= 1.0 ) then
	--	didModifyAccel = false  -- back to defaults
	--end
end

function AfterInit() 
print( "maxAccelX", maxAccelX, "minAccelX", minAccelX )
print( "maxAccelY", maxAccelY, "minAccelY", minAccelY )
print( "maxJerkXY", maxJerkXY, "minJerkXY", minJerkXY )	-- fix up acceleration and jerk, depending on the state
	maxAccelX = maxAccelX or DEFAULT_MAX_ACCEL
	minAccelX = minAccelX or DEFAULT_MIN_ACCEL
	maxAccelY = maxAccelX or DEFAULT_MAX_ACCEL
	minAccelY = minAccelX or DEFAULT_MIN_ACCEL
	maxJerkXY = maxJerkXY or DEFAULT_MAX_JERK
	minJerkXY = minJerkXY or DEFAULT_MIN_JERK
	
	
	if ( minJerkXY >= maxJerkXY ) then 
		minJerkXY = DEFAULT_MIN_JERK
	end
	
	if ( minAccelX >= maxAccelX ) then 
		minAccelX =  DEFAULT_MIN_ACCEL
	end
	
	if ( minAccelY >= maxAccelY ) then 
		minAccelY = DEFAULT_MIN_ACCEL
	end

print( "maxAccelX", maxAccelX, "minAccelX", minAccelX )
print( "maxAccelY", maxAccelY, "minAccelY", minAccelY )
print( "maxJerkXY", maxJerkXY, "minJerkXY", minJerkXY )
end

local currAccelPctFast = 1

function writeCurrAccelSettings()
	writeAccelSettings( currAccelPctFast );
end

local minLayerTime = 20
local maxAllowedFan = 1.0
local minUseableFan = .15

-- Our own prefix script has a line like: M201 X3000 Y3000
-- the easiest thing is to substitute it with numbers based on the quality settings
-- but if it's not found, we need to insert one.

--local didSparseThisLayer = false;    -- cleared at the beginning of every layer
--local currentLayerHeight = 0;		-- tracked as layers go by
local slowdownFactor = 1.0;

--print( "prevE" .. prevE )
-- TODO: reset slowdown on new layer start

for line in fin:lines() do
	local doEcho = true  -- by default we always write out the original line at the bottom of this loop, but it can be suppressed
	--fout:write( line .. "\n" );
	if ( line:match("^G1" ) ) then
		if ( line:match( MOVE4_PATTERN ) ) then
			-- parse it and remember the axis positions
			local x, y, z, e = parseGCodeMove4( line )
			if ( e == nil ) then
				print( "MOVE4_PATTERN e is nil " .. line .. " prevE " .. prevE )
			end
			prevX = x
			prevY = y
			prevZ = z
			prevE = e

		elseif ( line:match( MOVEXYEF_PATTERN ) ) then
			-- parse it and remember the axis positions, and grab the feedrate
			local x, y, e, f = parseGCodeMoveXYEF( line )
			prevX = x
			prevY = y
			local eDiff = e - prevE
			if ( eDiff > 0 ) then
				prevE = e
				-- it's an extruding move with feedrate set
				prevFeedrate = f
				
				-- also write the changed acceleration if that's appropriate
				if ( state == stateEmitAccelOnExtrudeMove ) then
					writeCurrAccelSettings();
					state = stateNull
				end
				
				
				--print( slowdownState )
				if ( slowdownFactor < 1 ) then
					-- we may modify the fedrate by feeding in more fan, up to a point
					local fanfraction = 0;  -- default to no fam
					local SLOWDOWN_FOR_FULL_FAN = .25 -- if slowdown hits this factor, use full fan.
					if ( slowdownFactor >= SLOWDOWN_FOR_FULL_FAN ) then
						fanfraction = 1 - ( slowdownFactor - SLOWDOWN_FOR_FULL_FAN ) / ( 1 - SLOWDOWN_FOR_FULL_FAN )
						-- modify the feed rate back up to full speed
						f = f / slowdownFactor  
						--print( fanfraction )
					else
						-- we're at the slowdown lower limit - below this the fan has to be full on, 
						-- and we will keep a slowdown factor so that the time is a fixed fraction of the
						-- set minimum layer time
						fanfraction = 1
						-- modify the speed based on a new effective slowdown factor as compared to the minimum
						-- so at minimum it's full speed, and below that it's proportional to the feed factor requested
						-- e.g. at 1/2 the slowdown for full fan it is e.g. .25 requested, now becomes .5 because it's .5 of /25
						f = f / SLOWDOWN_FOR_FULL_FAN
					end
					-- write out the modified line
					writeGivenXYEFLine( fout, x, y, e, math.ceil( f ) );
					if ( slowdownState == slowdownState_FixUpFan ) then
						local fan255 = linearInterpolatePct( fanfraction, minUseableFan, maxAllowedFan ) * 255
						writeFanControlLine( fout, math.ceil( fan255 ) );
					end
					doEcho = false;
					slowdownState = nil;
				end
			end
		elseif ( line:match( MOVE3_PATTERN ) ) then
			-- parse it and remember the axis positions
			local x, y, e = parseGCodeMove3( line )

			if ( e == nil ) then
				print( "MOVE3_PATTERN e is nil " .. line )
			end

			if ( state == stateEmitAccelOnExtrudeMove ) then
				--local xDiff = x - prevX
				--local yDiff = y - prevY
				local eDiff = e - prevE
				-- if we see a real extrusion move, first emit the latest modified aceleration settings
				-- do the real parse of the move, and see if it there is extrusion in the move.
				-- for this we must have parsed ALL G1 lines and kept track of the E position
				-- it could be a retract, but not likely. However we are looking for a positive extrusion anyway.
				-- It might include the Prime movement.
				if ( eDiff > 0 ) then
					-- it's an extrusion. We really don't need to check to see if there's an X,Y move too
					-- for it to be the point where we need to set the current accelerations
					writeCurrAccelSettings();
					state = stateNull
				end

			end
			prevX = x
			prevY = y
			prevE = e
			--print( "MOVE4_PATTERN prevE" .. prevE )

		elseif ( line:match( MOVE_E_PATTERN ) ) then
			-- parse it and remember the axis position
			prevE = parseGCodeMoveE( line )
			if ( prevE == nil ) then
				print( "MOVE_E_PATTERN e is nil" )
			end
		end

	elseif ( didModifyAccel and line:match( "^;$" ) ) then  -- single comment line - end of section

		--print "found end of section"

		-- restore the default (max) settings
		writeAccelSettings( 1.0 )  -- full speed settings
		doEcho = false
		
	elseif ( line:match( PERIMETER_PATTERN ) ) then

		--print "found Perimeter"
		local MAX_PERIM_PctFast = .5  -- clamp it for the perimeter - always limit the acceleration a bit
		local pctFast = 1.0 - quality
		pctFast = math.min( pctFast, MAX_PERIM_PctFast );
		currAccelPctFast = pctFast  -- save for when first real extrude move is found
		state = stateEmitAccelOnExtrudeMove
		--writeAccelSettings( pctFast )  -- modify the accelerations based on the stated quality setting
		
	elseif ( line:match( SLOWING_DOWN_PATTERN ) ) then
	
		slowdownFactor = parseSlowdownFactor( line )
		slowdownState = slowdownState_FixUpFan -- if we see a move with a feedrate... adjust it and make sure the fan is used instead up to a point
		
	elseif ( line:match( LOOP_PATTERN ) ) then

		--print "found Loop"

		-- a heuristic: half way to the quality setting?
		local pctFast = 1.0 - quality * .5  -- average between full speed and the quality setting speed

		currAccelPctFast = pctFast  -- save for when first real extrude move is found
		state = stateEmitAccelOnExtrudeMove

	elseif ( line:match( BEGIN_LAYER_PATTERN ) ) then
		slowdownFactor = 1.0 -- reset every layer in case it's not there
	elseif ( line:match( "Speed vs Quality" ) ) then

		--print "found quality setting"
		-- parse the setting
		quality = getNumericSettingValue( line )
		print( "Quality " .. quality )

	elseif ( line:match( "^M201" ) ) then     -- must begin the line, or it might be commented out

		--print "found max Acceleration settings"
		-- parse the setting

		local gtype, codeNum, axis1, axis1val, axis2, axis2val = parseGCodeSetting2Value( line )
		--print( "Acceleration: " .. gtype .. codeNum .. axis1 .. axis1val .. axis2 .. axis2val )
		--print( "Acceleration: " .. makeGCode2Value( gtype, codeNum, axis1, axis1val, axis2, axis2val ) )
		if ( type ) then
			foundAccelSpec = true
			if ( axis1 == 'X' and axis2 == 'Y' ) then
				if ( minAccelX == nil or axis1val < minAccelX ) then minAccelX = axis1val end
				if ( maxAccelX == nil or axis1val > maxAccelX ) then maxAccelX = axis1val end
				-- axis 2 is Y
				if ( maxAccelY == nil or axis2val > maxAccelY ) then maxAccelY = axis2val end
				if ( minAccelY == nil or axis2val < minAccelY ) then minAccelY = axis2val end
				doEcho = false
			end
		end
	elseif ( line:match( "^M205" ) ) then     -- must begin the line, or it might be commented out

		--print "found max Jerk settings"
		-- parse the setting

		local gtype, codeNum, axis1, axis1val = parseGCodeSetting1Value( line )
		print( "Jerk: " .. makeGCode1Value( gtype, codeNum, axis1, axis1val ) )
		--print( "Jerk: " .. makeGCode1Value( gtype, codeNum, axis1, axis1val, axis2, axis2val ) )
		if ( type ) then
			foundJerkSpec = true
			if ( axis1 == 'X' ) then
				if ( maxJerkXY == nil or axis1val > maxJerkXY ) then maxJerkXY = axis1val end
				if ( minJerkXY == nil or axis1val < minJerkXY ) then minJerkXY = axis1val end
				doEcho = false
			end
		end

	elseif ( line:match( "Main%sG%-code" ) ) then
		
		AfterInit() 			
		--print "found Main G-Code"
		--whether we have found the spec or not, write out a line that sets it, based on the quality setting
		writeAccelSettings( 1.0 )  -- full speed settings to begin with
		inMainGCode = true

		--writeAccelSettings( pctFast )

--[[	elseif ( line:match( SPARSEFILL_PATTERN ) ) then

		--print "found sparse infill"
		fout:write( line .. "\n" )
		-- at very high speeds, the sparse infill parallel lines that are drawn to bridge
		-- above the stcaked infill do not stick well at all.
		-- We may need to slow them down, but first try just using less acceleration.
		local MAX_SPARSE_PctFast = .75  -- just clamp it a bit for the first sparse infill above stacked infill
		--local MAX_SPARSE_PctFast = .1  -- clamp it for the first sparse infill above stacked infill
		local pctFast = 1.0 - quality
		pctFast = math.min( pctFast, MAX_SPARSE_PctFast );
		writeAccelSettings( pctFast )  -- modify the accelerations
]]--
	
	end

	if ( doEcho == true ) then
		fout:write( line, "\n" )  -- echo it out
	end

end

-- done writing
fin:close();
fout:close();

if ( replaceOriginal == true ) then
	-- rename the original file
	local newOriginalName = inFileName .. "x"     -- just change the extension
	os.remove( newOriginalName );                  -- get rid of any older temporary file if there is one
	os.rename( inFileName, newOriginalName );  -- rename the original to the new name
	os.rename( outFileName, inFileName );        -- and replace original with processed one
end

print "done"

if ( replaceOriginal ~= true ) then
	print(string.format("elapsed time: %.2f\n", os.clock() - x))
	local answer
	--answer = io.read()
end
--answer = io.read()



